import numpy as np
cimport numpy as np
from libc.stdlib cimport malloc, free
from libc.string cimport memset

cdef extern from "math.h":
    double sqrt(double theta)


cdef pldist(double point[3], double start[3], double end[3]):
    """
    Calculates the distance from ``point`` to the line given
    by the points ``start`` and ``end``.
    """
    cdef double res, proj_dist
    cdef double direction[3], vec[3], proj[3]
    cdef double line_dist
    if start[0] == end[0] and start[1] == end[1] and start[2] == end[2]:
        res = sqrt((point[0] - start[0]) * (point[0] - start[0]) + (point[1] - start[1]) * (point[1] - start[1]) + (point[2] - start[2]) * (point[2] - start[2]))
        return res
    line_dist = sqrt((start[0] - end[0]) * (start[0] - end[0]) + (start[1] - end[1]) * (start[1] - end[1]) + (start[2] - end[2]) * (start[2] - end[2]))
    direction[0] = (end[0] - start[0])/line_dist
    direction[1] = (end[1] - start[1])/line_dist
    direction[2] = (end[2] - start[2])/line_dist
    vec[0] = point[0] - start[0]
    vec[1] = point[1] - start[1]
    vec[2] = point[2] - start[2]
    proj_dist = vec[0] * direction[0] + vec[1] * direction[1] + vec[2] * direction[2]
    proj[0] = start[0] + proj_dist * direction[0]
    proj[1] = start[1] + proj_dist * direction[1]
    proj[2] = start[2] + proj_dist * direction[2]
    res = sqrt((proj[0] - point[0]) * (proj[0] - point[0]) + (proj[1] - point[1]) * (proj[1] - point[1]) + (proj[2] - point[2]) * (proj[2] - point[2]))
    return res

cdef _c_rdp(points, int n, char* mask, double epsilon):
    stk = [(0, n - 1)]
    cdef int st, ed, index
    cdef double dmax #, p0, p1, p2,dis 
    cdef double start[3], end[3], point[3]
    while stk:
        st, ed = stk.pop()
        start = points[st]
        end = points[ed]
        dmax = 0.0
        index = st
        for i from st < i < ed:
            point = points[i]
            d = pldist(point, start, end)
            if d > dmax:
                index = i
                dmax = d
        if dmax > epsilon:
            stk.append((st, index))
            stk.append((index, ed))
        else:
            for i from st < i < ed:
                mask[i] = 0

def rdp(points_in, double epsilon=0):
    cdef int n
    n = len(points_in)
    cdef np.ndarray[np.float64_t, ndim=2] points = np.asanyarray(points_in)
    is_2d = False
    if len(points[0]) == 2:
        is_2d = True
        dummy_z = np.full(n, 0.0)
        points = np.column_stack((points, dummy_z))
    cdef char *mask = <char*> malloc(n)
    memset(mask, 1, n)
    _c_rdp(points, n, mask, epsilon)
    res = []
    for i from 0<=i<n:
        if mask[i]:
            res.append(points[i])
    free(mask)
    np_res = np.asanyarray(res)
    if is_2d:
        np_res = np.column_stack((np_res[:, 0], np_res[:, 1]))
    return np_res
